
( BADNOP dev env for PICmicro 18fxx2 )
( ********** 18f452 FORTH MACROS  ************ )
(

What you find here is the code generator for the 18fxx2 architecture.
All words ending with a ',' character are by convention words that generate
code at the current assembly point. They are invoked whenever a target
word's compilation encounters the name without the ',' character.

The compiler's data stack can be used to store syntactic data about
the word being compiled, like for instance is used in the implementation
of the for..next and if..then constructs.

Note that target's ':' word cannot load literals on the compiler stack.

)



( symbolic constants )
load 18f452.f


( INSTRUCTION SET )

( 18fxx2 opcodes )

: andwf  0x1400 ;
: addwf  0x2400 ;
: subwf  0x5C00 ;
: decf   0x0400 ;
: incf   0x0400 ;  ( FIX ME ! )
: xorwf  0x1800 ;
: iorwf  0x1000 ;

: movwf  0x6E00 ;
: movf   0x5000 ;
: movlw  0x0E00 ;
: movff  0xC000 ;

: addlw  0x0F00 ;
: sublw  0x0800 ;
: andlw  0x0B00 ;
: xorlw  0x0A00 ;
: iorlw  0x0900 ;


: push   0x0005 ;
: pop    0x0006 ;

: bz     0xE000 ;
: bov    0xE400 ;
: bnov   0xE500 ;
: bnn    0xE700 ;
: bnc    0xE300 ;
: bn     0xE600 ;
: bc     0xE200 ;
: bnz    0xE100 ;
: bra    0xD000 ;
: rcall  0xD800 ;

: nop    0x0000 ;
: nopf   0xf000 ;   ( nop used as extension word for 2 word ins )
: return 0x0012 ;
: retfie 0x0010 ;
: call   0xEC00 ;
: goto   0xEF00 ;

: lfsr   0xEE00 ;


( instruction opcode modifiers : inclusive or masks )
: w/   ;                     ( destination wreg DEFAULT )
: reg/      0x0200 or ;      ( destination freg )
: access/ ;                  ( access bank DEFAULT )
: banked/   0x0100 or ;      ( bsr bank )
: op/       0xff and or ;    ( add operand to instruction )
: sop/      swap op/ ;

( forth opcodes  )
: op-drop movf  POSTDEC0 op/ w/ ;
: op-dup  movwf PREINC0  op/ reg/ ;

( peephole optimization for certain instructions:
-- stack movement optimization: makes sense since usually dup/drop is an explicit instruction
-- tail call optimisation: necessary for proper usage of tail recursion
-- literal folding. more so: ! and @ only support literal arguments

the optimizer is very simple: just emit code for each word in a colon definition
if there is no possible optimization, else undo a previous compilation and combine
it's effect with the current one. )

( instruction undo stacks )
stack last-dup
stack last-drop
stack last-lit
stack last-call

: record point swap stack-push ;        ( record current point on stack )
: top=last? ( stack -- T/F )
    stack-top point 2 - = ;		( compare top of peephole stack to previous point )
: top=lastbig? ( stack -- T/F ) 
    stack-top point 4 - = ;             ( same but for 2word instructions )

: opti-reset      ( turn off peephole optimization
                    this needs to be called at jump targets: i.e. next and docolon )
    last-dup  stack-reset
    last-drop stack-reset
    last-lit  stack-reset
    last-call stack-reset ;

: :start opti-reset ;   ( start target colon word hook )


( MACROS - map forth to assembler )

( utilities and small instructions )
: undo    ( undo compilation of last instruction )
    point 2 - !point ;
: movff,  ( src dst -- )
    swap 0xfff and movff or ,
    0xfff and nopf or , ;
: lfsr,   ( value reg --  )
    >r dup
    8 >> 0x0f and lfsr r> 4 << or or , ( compile first word )
    0xff and nopf or , ;               ( compile second word )

: nop,  nop , ;
: push, push , ;
: pop,  pop , ;


( nop messages )
: bad,     point 0xfbad , 0xbad .x .x ;



( stack manips )
: silentdrop,    POSTDEC0 WREG movff, ;   ( does not affect flags )
: test,          andlw 0xff or , ;        ( test Z and N flags from word on tos )
: >flags,        test, silentdrop, ;
    
    
( drop/save and opti )
: dup,       last-dup  record op-dup  , ; ( record last dup )
: got-dup?   last-dup  top=last? ;
: undo-dup   undo last-dup stack-drop ;

: drop,
    got-dup? pass undo-dup
    last-drop record op-drop , ;       ( else record and compile )
: got-drop?  last-drop top=last? ;
: undo-drop  undo last-drop stack-drop ;

: save,      got-drop? not pass dup, undo-drop ;



( jump/call/ret and opti )
: got-call?     last-call top=lastbig? ;    ( check if last instruction was a call )
: tailcall,  ( turn last call into a tail call )
    point 4 - p@ 
    0xfff and goto or
    point 4 - p!
    last-call stack-drop ;
: ;,
    got-call? pass tailcall, return , ; ( compile a tail call or a return )
: ;;, ;, ;  ( fog )
: blong,    ( operands opcode -- ) ( compile a 2word long branch instruction )
    >r 2 / dup                     ( save opcode and convert to word addressing )
    0xff and r> or ,
    8 >> 0xff and nopf or , ;
: call,   last-call record call blong, ;
: jump,   goto blong, ;



( literal and optimization
there are only 2 possible code transformations:
/dup lit/ -> //             undo the lit completely
/lit/     -> /drop/         undo an optimized lit
                            the drop can be undone by a following save, )
: lit,
    save,
    last-lit record                ( save point for opti )
    movlw sop/ ,  ;
: lit2,                            ( , two byte literal )
    dup lit, 8 >> lit, ;

: get-lit                          ( get literal from last instruction )
    point 2 - p@ 0xff and ;

: got-lit?  last-lit top=last? ;   ( last instruction a literal? )

: undo-lit
    undo last-lit stack-drop                  ( undo movlw )
    got-dup? pass undo-dup	              ( "dup lit" -> "" )
    drop, ;                                   ( "lit" -> "drop" )
: not-lit?     got-lit? not ;
: re-lit       get-lit undo-lit ;             ( get literal and undo literal compilation )

load 18f452_mac.f   ( load symbolic name macros after lit, is defined )


( set / store )
: litstore,    movwf sop/ , ;
: lit!,        litstore, drop, ;

: !,
	not-lit? pass bad,	( prev inst not a literal? -> not supported )
	re-lit  		( load lit value on DS and undo lit instruction )
	got-dup? not pass lit!, ( no dup -> compile a store+drop )
	undo-dup litstore, ;    ( dup -> undo dup and compile store )



( get / fetch )
: litload,     movf sop/ , ;
: lit@,        save, litload,  ;
: @,           not-lit? pass bad, re-lit save, litload, ;



( TOS words )
: neg,      sublw , ;
: not,      xorlw 0xff op/ , ;

( 2 argument words  )
: 2op,      POSTDEC0 op/ reg/ , ;      ( compile stack 2op instruction )

: real+,    addwf 2op, ;
: real-,    subwf 2op, ;   
: realxor,  xorwf 2op, ;
: realand,  andwf 2op, ;
: realor,   iorwf 2op, ;

( arithmic operating on TOS and immediate )
: lit+,     addlw sop/ , ;               ( compile literal add )
: lit-,     neg lit+, ;                  ( literal sub )
: litxor,   xorlw sop/ , ;
: litand,   andlw sop/ , ;
: litor,    iorlw sop/ , ;


( literal folding )
: +,        not-lit?  pass real+,    re-lit lit+, ;
: -,        not-lit?  pass real-,    re-lit lit-, ;
: or,       not-lit?  pass realor,   re-lit litor, ;
: xor,      not-lit?  pass realxor,  re-lit litxor, ;
: and,      not-lit?  pass realand,  re-lit litand, ;

: incf,     incf sop/ reg/ , ;
: decf,     decf sop/ reg/ , ;

( return stack 'local variables' )
: rl,       TOSL, ;
: rh,       TOSH, ;

( for counters )
: rl++,     TOSL incf, ;
: rl--,     TOSL decf, ;
: rh++,     TOSH incf, ;
: rh--,     TOSH decf, ;

( push/pop might block opti?? )
: here,   push, rl--, rl--, ;               ( push location )
: >r,     push, rl, !, ;                    ( move byte to/from return stack )
: r>,     rl, @, pop, ;
: >>r,    push, rh, !, rl, !, ;             ( move word to/from return stack )
: r>>,    rl, @, rh, @, pop, ;
: rdrop,  pop, ;


: TMP 0x10 ;  ( temporary register. needs to be saved during interrupts )
: TMP, TMP lit, ;

: swap,                                     ( 4 cycles.. )
    INDF0 TMP movff,
    movwf INDF0 op/ ,
    movf  TMP op/ , ;





( conditional blocks use the machine flags )

: encode-srcdst   ( to from -- jmpoperand )
    -             ( distance )
    2 -           ( - 2 to compensate for PC inc. )
    2 / ;         ( instruction uses word addressing )
: offset-to    point encode-srcdst  ;              ( compute branch offset point->address )
: offset-from  point swap encode-srcdst ;          (                       address->point )
: rjump,       offset-to 0x7ff and bra or , ;      ( short relative jump )
: rcall,       offset-to  0x7ff and rcall or , ;   ( short relative call )
: op-cjump
    ( address opcode -- packedopcode ) ( create conditional jump opcode )
    >r offset-to 0xff and r> or ;
: patch-cjump, ( addr )
    dup offset-from 0xff and >r  ( compute relative distance ) 
    dup p@                       ( get opcode )
    0xff00 and r> or             ( strip address and insert new )
    swap p! ;                    ( store instruction )
: for,
    >r, point        ( move count to r, save point on compiler stack )
    opti-reset ;       ( disable peephole opti )
: next,              ( decrement value in TOSL and branch back if nonzero )
    rl--,
    bnz op-cjump ,
    pop , ;
: ifnz,  point bz , ;
: ifz,   point bnz , ;
: ifc,   point bnc , ;
: ifnc,  point bc , ;
: ifn,   point bnn , ;
: ifnn,  point bn , ;
: ifov,  point bnov , ;
: ifnov, point bov , ;
: skip,  point bra , ;
: if,    >flags, ifnz, ;   ( proper if uses tos. use flag ifs if possible. they are faster )
: then,                    ( can be used to patch any conditional reljump )
    patch-cjump,           ( it can patch forward bra too )
    opti-reset ;
: else,                    ( patch first cond jump and set new uncond )
    skip, >r patch-cjump, r> ;



( indirect addressing uses FSR1,FSR2 : the 'a' and 'b' registers )

( changing pointer )
: lit!a, 1 lfsr, ;                ( literal set a )
: lit!b, 2 lfsr, ;                ( literal set b )

: al, FSR1L, ;
: ah, FSR1H, ;
: bl, FSR2L, ;
: bh, FSR2H, ;

( save/restore a/b to return stack )
: a>r,  push, al, @, rl, !, ah, @, rh, !, ;
: b>r,  push, bl, @, rl, !, bh, @, rh, !, ;
: r>a,  rh, @, al, !, rl, @, ah, !, pop, ;
: r>b,  rh, @, bl, !, rl, @, bh, !, pop, ;

( register use: compiles literal for deref operation )
: a,     INDF1, ;
: a+,    POSTINC1, ;
: +a,    PREINC1, ;
: a-,    POSTDEC1, ;

: b,     INDF2, ;
: b+,    POSTINC2, ;
: +b,    PREINC2, ;
: b-,    POSTDEC2, ;


( interrupt handlers )

: install-reset ( startword -- )
    point >r ( save point )

    ( compile relative jump to start routine )
    0x0 !point      ( compile at 0x0 )
    0x40 rjump,     ( jump to 0x40 )

    ( init routine: setup forth environment )
    0x40 !point         ( compile at 0x40 )
    movlw ,             ( clear TOS )
    0x20 0 lfsr,        ( init data stack ptr )
    movwf STKPTR op/ ,  ( init return stack ptr )
    rjump,              ( jump to start word )
           
    r> !point ; ( restore point )

: install-ihigh ( intword -- )
    point >r

    0x8 !point          ( move point to iv )
    TMP, @,             ( thanks to swap )
    call,               ( call handling word )
    TMP, !,
    retfie 1 or ,       ( fast return )
    
    r> !point ;

: install-ilow ( intword -- ) 
    point >r
    0x18 !point

    STATUS, @,          ( save flags )
    TMP, @,
    call,
    TMP, !,
    STATUS, !,	        ( restore flags )
    retfie ,

    r> !point ;



( simulator - none yet
simple one: just run the instructions inside the mother forth
using mother data & return stack
)

var target-next

: target-next++  ;

: do-return ;
: do-jump ;
: do-call ;

: target-execute .x ; ( just echo address )
    



0x80 !point         ( start compiling forth code at 0x80 ) 


target

( host communication.
execute word
block program
)

( uart stuff )


( port config )

( timer )

host
